/* cswplay.c
 *
 * Copyright (C) 2005 Darren Salt
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <fcntl.h>
#include <errno.h>
#include <zlib.h>

typedef unsigned char byte;

static int audio;

struct csw_t {
  int rate;
  long pulses;
  byte polarity;
  byte compression;
};

static int
read_n (int fd, void *buf, size_t count)
{
  errno = 0;
  return read (fd, buf, count) != count;
}

static int
check_header (const char *const name, int fd, struct csw_t csw[])
{
  byte buf[256];

  if (read_n (fd, buf, 23))
    goto err;
  if (memcmp (buf, "Compressed Square Wave\x1A", 23))
  {
    fprintf (stderr, "cswplay: %s: not a CSW file\n", name);
    return 1;
  }

  if (read_n (fd, buf, 2))
    goto err;
  switch (buf[0])
  {
  case 1:
    if (read (fd, buf, 7))
      goto err;
    csw->rate = buf[0] | buf[1] << 8;
    if (csw->rate < 1)
      goto bad_header;
    csw->pulses = -1;
    if (buf[2] != 1)
      goto bad_header;
    csw->compression = 0;
    csw->polarity = buf[3] & 1;
    break;

  case 2:
    if (read_n (fd, buf, 27))
      goto err;
    csw->rate = buf[0] | buf[1] << 8 | buf[2] << 16 | buf[3] << 24;
    if (csw->rate < 1 || csw->rate >= 0x80000)
      goto bad_header;
    csw->pulses = buf[4] | buf[5] << 8 | buf[6] << 16 | buf[7] << 24;
    if (buf[8] < 1 || buf[8] > 2)
      goto bad_header;
    csw->compression = buf[8] - 1;
    csw->polarity = buf[9] & 1;
    {
      int ext = buf[10];
      if (ext && read_n (fd, buf, ext))
        goto err;
    }
    break;

  default:
    fprintf (stderr, "cswplay: %s: unrecognised version %d.%d\n",
	     name, buf[0], buf[1]);
    return 1;
  }

  return 0;

  err:
  fprintf (stderr, "cswplay: %s: %s\n", name,
	   errno ? strerror (errno) : "unexpected end-of-file");
  return 1;

  bad_header:
  fprintf (stderr, "cswplay: %s: invalid header data\n", name);
  return 1;
}

static int
playcsw (const char *const name, int fd)
{
  struct csw_t csw;
  byte buf[65536], in[65536], sample[65536];
  size_t bptr, sptr, index, count;
  long pulse;

  z_stream gz;
  int zinit = 0;

  if (check_header (name, fd, &csw))
    return 1;

  /* set the sample rate if we're not writing to stdout */
  if (ioctl (audio, SNDCTL_DSP_SPEED, &csw.rate))
    if (audio != 1)
      fprintf (stderr, "cswplay: warning: cannot set sample rate to %d\n",
	       csw.rate);

  gz.next_in = in;
  gz.avail_in = 0;

  index = sptr = count = pulse = 0;
  bptr = -1;
  for (;;) /* while (csw.pulses) */
  {
    if (!count)
    {
      bptr = -1;
      if (csw.compression)
      {
	if (gz.avail_in)
	  memmove (in, gz.next_in, gz.avail_in);

	count = read (fd, gz.next_in + gz.avail_in, sizeof (in) - gz.avail_in);
	if (count < 0)
	  goto zerror;

	gz.next_in = in;
	gz.avail_in += count;
	gz.next_out = buf;
	gz.avail_out = sizeof (buf);

	if (!zinit)
	{
	  zinit = 1;
	  gz.zalloc = Z_NULL;
	  gz.zfree = Z_NULL;
	  gz.opaque = Z_NULL;
	  if (inflateInit (&gz) != Z_OK)
	    goto zerror;
	}

	switch (inflate (&gz, 0))
	{
	case Z_STREAM_END:
	case Z_OK:
	  break;
	default:
	  goto zerror;
	}
	count = gz.next_out - buf;
      }
      else
	count = read (fd, buf, sizeof (buf));
      if (count == 0)
	break;
      if (count < 0)
        goto zerror;
    }

    --count;
    ++bptr;
    if (index)
    {
      pulse |= buf[bptr] << (index - 1);
      index += 8;
      if (index >= 32)
	index = 0;
    }
    else if (buf[bptr])
      pulse = buf[bptr];
    else
    {
      pulse = 0;
      index = 1;
      continue;
    }

    if (pulse < 0 || pulse > INT_MAX)
      pulse = INT_MAX;

    while (pulse)
    {
      long chunk = sizeof (sample) - sptr;
      if (chunk > pulse)
	chunk = pulse;
      memset (sample + sptr, csw.polarity ? 0x90 : 0x70, chunk);
      sptr += chunk;
      pulse -= chunk;
      if (sptr == sizeof (sample))
      {
	write (audio, sample, sptr);
	sptr = 0;
      }
    }
    csw.polarity ^= 1;

    if (csw.pulses != -1)
      --csw.pulses;
  }

  if (sptr)
    write (audio, sample, sptr);

  ioctl (audio, SNDCTL_DSP_SYNC, NULL);
  return 0;

  zerror:
  fprintf (stderr, "cswplay: %s: %s\n", name,
	   (csw.compression && gz.msg) ? gz.msg : strerror (errno));

  ioctl (audio, SNDCTL_DSP_SYNC, NULL);
  return 1;
}

int
main (int argc, char **argv)
{
  int i, ret = 0;

  if (isatty (1))
  {
    audio = open ("/dev/dsp", O_WRONLY);
    if (audio == -1)
      goto audio_error;
  }
  else
    audio = 1; // should be stdout

  i = AFMT_U8;
  if (ioctl (audio, SNDCTL_DSP_SETFMT, &i)) // want 8-bit unsigned
    if (audio != 1)
      goto audio_error;
  i = 0;
  if (ioctl (audio, SNDCTL_DSP_STEREO, &i)) // want mono
    if (audio != 1)
      goto audio_error;

  if (argc == 1)
    return playcsw ("-", 0 /* stdin */);

  for (i = 1; i < argc; ++i)
  {
    int fd = open (argv[i], O_RDONLY | O_NOCTTY);
    if (fd >= 0)
    {
      ret |= playcsw (argv[i], fd);
      close (fd);
    }
    else
    {
      fprintf (stderr, "cswplay: %s: %s\n",
	       argv[i], strerror (errno));
      ret = 1;
    }
  }

  close (audio);
  return ret;

  audio_error:
  fprintf (stderr, "cswplay: %s: %s\n", "/dev/dsp", strerror (errno));
  close (audio);
  return 1;
}
